/******************************************************************************
*
* project name:    TI-Chess
* file name:       loadsave.c
* initial date:    19/03/2000
* authors:         thomas.nussbaumer@gmx.net (coding)
*                  marcos.lopez@gmx.net      (design/graphics/beta testing)
* description:     contains load/save routines
*
* $Id: loadsave.c,v 1.5 2001/02/17 15:00:11 Thomas Nussbaumer Exp $
*
******************************************************************************/


short problem_active = 0;
short problem_max    = 0;

char problem_text1[SIZE_TEXT+1] = {0};
char problem_text2[SIZE_TEXT+1] = {0};
char problem_text3[SIZE_TEXT+1] = {0};
char problem_copyright[SIZE_COPYRIGHT+1] = {0};



//-----------------------------------------------------------------------------
// declarations of the necessary "external" variables
//-----------------------------------------------------------------------------
extern short          updatecounter;
extern short          already_mate;
extern char*          board;
extern short*         ep_field;
extern short*         stack_ranges;
extern short*         material_bal;
extern short*         material_sum;
extern short          white_king_pos;
extern short          black_king_pos;
extern short          act_color;
extern short          act_depth;
extern long           nr_nodes;
extern unsigned short move_count;
extern short          moves_stored;
extern short          moves_forward;
extern unsigned short flags[MAX_DEPTH];
extern defaults_t     defaults;
extern store_t*       move_store;
extern const char     fig_symbols[NR_FIGS*2-1];
extern const short    fig_values[NR_FIGS];
extern short          loadsave_idx;
extern short          twoplayer_mode;


/*===========================================================================*/
/* returns TRUE if given file is archived                                    */
/*===========================================================================*/
short isArchived(char* name) {
    SYM_ENTRY* symentry = DerefSym(SymFind(name));

    if (!symentry) return FALSE;

    return (symentry->flags.bits.archived || symentry->flags.bits.locked);
}



/*===========================================================================*/
/* if automatic == TRUE: uses ticsave0 for loading                           */
/* else                  uses name from load/save menu                       */
/*===========================================================================*/
short HandleLoad(short automatic,short load_problem) {
    SYM_ENTRY*     symentry;
    unsigned char* varptr;
    short          i;
    short          idx;
    unsigned char  lo;
    unsigned char  hi;
    short          content;
    unsigned short abs_val;
    short          moves = -1;

    unsigned char  use = 0;

    if (load_problem) {
        symentry = SymFindPtr ($(ticpro), 0);
    }
    else {
        if (!automatic) use = loadsave_idx;

        switch(use) {
            case 1:  symentry = SymFindPtr ($(ticsave1), 0); break;
            case 2:  symentry = SymFindPtr ($(ticsave2), 0); break;
            case 3:  symentry = SymFindPtr ($(ticsave3), 0); break;
            default: symentry = SymFindPtr ($(ticsave0), 0); break;
        }
    }

    if (!symentry) return FALSE;

    varptr = (unsigned char*)HeapDeref(symentry->handle);

    if (!varptr) return FALSE;


    if (!load_problem) {
        moves = ((short)varptr[0]) * 256 + ((short)varptr[1]) - 39;
        if (moves % 3) return FALSE;
        moves /= 3;
    }
    else {
        varptr += SIZE_PROBLEM*(problem_active - 1) + SIZE_TEXT*3+SIZE_COPYRIGHT + 2;
    }

    idx = A1;
    for (i=3;i<35;i++) {
        hi = (varptr[i] & 0xf0) >> 4;
        lo = (varptr[i] & 0x0f);

        if (hi == W_KING)         white_king_pos = idx;
        else if (hi == B_KING+13) black_king_pos = idx;

        if (lo == W_KING)         white_king_pos = idx+1;
        else if (lo == B_KING+13) black_king_pos = idx+1;

        if (hi > 6) board[idx] = ((char)hi) - 13;
        else        board[idx] = ((char)hi);
        idx++;
        if (lo > 6) board[idx] = ((char)lo) - 13;
        else        board[idx] = ((char)lo);
        idx++;
        if (idx%10==9) idx+=2;
    }

    move_count      = (unsigned short) ((varptr[35] << 8)| varptr[36]);
    flags[0]        = (unsigned short) ((varptr[37] << 8)| varptr[38]);

    if ((move_count % 2) == 0) act_color = WHITE;
    else                       act_color = BLACK;

    material_sum[0] = 0;
    material_bal[0] = 0;

    for (i=A1;i<=H8;i++) {
       content = board[i];

       if (content == OUTSIDE || content == EMPTY) continue;
       abs_val = abs(content);
       if (content < 0) {
           material_bal[0] -= fig_values[abs_val];
       }
       else {
           material_bal[0] += fig_values[abs_val];
       }

       if (abs_val == W_PAWN) continue;
       material_sum[0] += fig_values[abs_val];
    }

    if (act_color == BLACK) material_bal[0] = -material_bal[0];

    // UpdateNodeDisplay(0);

    already_mate    = FALSE;
    updatecounter   = 0;
    moves_stored    = 0;
    act_depth       = 0;
    ep_field[0]     = ILLEGAL;
    stack_ranges[0] = 0;
    nr_nodes        = 0;

    CalcHashKey();
    InitPawnVars();
    ResetHashCounters();

    if (moves > 0) {
        char m_t[6];
        char from;
        char to;
        char promote;
        for (i=0;i<moves;i++) {
            from    = (char)varptr[39+i*3];
            to      = (char)varptr[39+i*3+1];
            promote = (char)varptr[39+i*3+2];

            Field2Notation((short)from,m_t);
            Field2Notation((short)to,&m_t[2]);
            if (promote != EMPTY) {
                m_t[4] = fig_symbols[promote+6];
                m_t[5] = 0;
            }
            else {
                m_t[4] = 0;
            }

            if (i < moves-1) UserMove(m_t,FALSE);
            else {
                twoplayer_mode = TRUE;
                UserMove(m_t,TRUE);
                twoplayer_mode = FALSE;
            }
        }
    }

    return TRUE;
}



/*===========================================================================*/
/* if auto == TRUE: uses ticsave0 for saving                                 */
/* else             uses savename from loadsave menu                         */
/*===========================================================================*/
short HandleSave(short automatic) {
    SYM_ENTRY*     symentry;
    unsigned char* varptr;
    short          i;
    short          idx;
    char           lo,hi;
    char           use   = 3;
    short          moves = 0;
    char*          name;

    if (!automatic) use = loadsave_idx - 4;

    switch(use) {
        case 0:  name = $(ticsave1); break;
        case 1:  name = $(ticsave2); break;
        case 2:  name = $(ticsave3); break;
        default: name = $(ticsave0); break;
    }

    if (isArchived(name)) return FALSE;

    symentry = DerefSym(SymAdd(name));

    if (!symentry) return FALSE;

    if (defaults.savemoves == OPTION_ON) {
        moves = moves_stored;
        varptr = (unsigned char*)HeapDeref(symentry->handle = HeapAlloc(40+5+moves*3));
        idx = 36+3+moves*3;
    }
    else {
        varptr = (unsigned char*)HeapDeref(symentry->handle = HeapAlloc(40+5));
        idx = 36+3;
    }

    if (!varptr) return FALSE;

    varptr[0] = (unsigned char)(idx/256);  // 2 bytes: length of the variable - 2
    varptr[1] = (unsigned char)(idx%256);
    varptr[2] = 0;    // zero marks a begining of the actual variable data

    if (defaults.savemoves == OPTION_ON) {
        // take back all stored moves and save initial position
        while (moves_stored > 0) TakeBackMove(TRUE);
    }

    idx = 3;
    for (i=A1;i<H8;i+=2) {
        if (board[i] == OUTSIDE) continue;
        hi = board[i];
        lo = board[i+1];
        if (lo<0) lo += 13;
        if (hi<0) hi += 13;
        varptr[idx] = (hi & 0x0f) << 4;
        varptr[idx] |= (lo & 0x0f);
        idx++;
    }

    varptr[idx++] = (move_count & 0xff00) >> 8;
    varptr[idx++] = move_count & 0xff;
    varptr[idx++] = (flags[0] & 0xff00) >> 8;
    varptr[idx++] = flags[0] & 0xff;

    if (defaults.savemoves == OPTION_ON) {
        // forward taken moves again and save moves
        while (moves > 0) {
            UndoTakeBackMove(TRUE);
            moves--;
        }

        for (i=0;i<moves_stored;i++) {
            varptr[idx++] = (unsigned char)move_store[i].from;
            varptr[idx++] = (unsigned char)move_store[i].to;
            varptr[idx++] = (unsigned char)move_store[i].promoted_fig;
        }
    }

    varptr[idx++] = 0;    // end-of-string marker
    varptr[idx]   = 0x2D; // the last byte is a type (0x2D = STR)

    return TRUE;
}


defaults_t default_defaults = { STD_USERLEVEL-1, // strength
                                OPTION_ON,       // autoload
                                OPTION_ON,       // autosave
                                OPTION_ON,       // hashtables
                                OPTION_ON,       // firsttime
                                OPTION_ON,       // requests
                                OPTION_ON,       // save moves
                                OPTION_ON,       // default pieceset
                                OPTION_ON};      // unused



/*===========================================================================*/
/* load defaults from String variable "ticcfg"                               */
/*===========================================================================*/
short LoadDefaults(void) {
    FILE *fp;
    short result;

    if (!(fp = fopen("ticcfg","rb"))) {
        memcpy(&defaults,&default_defaults,sizeof(defaults_t));
        SaveDefaults();
        return FALSE;
    }

    result = fread(&defaults,sizeof(defaults_t),1,fp);
    fclose(fp);

    if (result != 1) {
        memcpy(&defaults,&default_defaults,sizeof(defaults_t));
        SaveDefaults();
    }

    return result;
}



/*===========================================================================*/
/* save defaults to String variable "ticcfg"                                 */
/*===========================================================================*/
short SaveDefaults(void) {
    FILE *fp;
    unsigned short result;

    EM_moveSymFromExtMem($(ticcfg), HS_NULL);
    if (!(fp = fopen("ticcfg","wb"))) return FALSE;
    result = fwrite(&defaults,sizeof(defaults_t),1,fp);
    fputc(0,fp);
    fputc(0x2D,fp);
    fclose(fp);
    return result;
}



/*===========================================================================*/
/* will load description of problem                                          */
/*===========================================================================*/
void LoadProblemDesc(void) {
    SYM_ENTRY*     symentry = SymFindPtr ($(ticpro), 0);
    unsigned char* varptr;
    unsigned short tmp;

    if (!symentry) {
        problem_max = -1;
        return;
    }

    varptr = (unsigned char*)HeapDeref(symentry->handle);
    if (!varptr) {
        problem_max = -1;
        return;
    }

    varptr+=3;

    tmp = *varptr << 8;
    varptr++;
    tmp += *varptr;
    problem_max = (short)tmp;
    varptr++;

    if (problem_max == 0) {
        problem_max = -1;
        return;
    }

    if (problem_active > problem_max) problem_active = problem_max;
    else if (problem_active <= 0) problem_active = 1;

    varptr += SIZE_PROBLEM*(problem_active-1);

    memset(problem_text1,0,SIZE_TEXT+1);
    memset(problem_text2,0,SIZE_TEXT+1);
    memset(problem_text3,0,SIZE_TEXT+1);
    memset(problem_copyright,0,SIZE_COPYRIGHT+1);

    strncpy(problem_text1,varptr,SIZE_TEXT);
    varptr += SIZE_TEXT;
    strncpy(problem_text2,varptr,SIZE_TEXT);
    varptr += SIZE_TEXT;
    strncpy(problem_text3,varptr,SIZE_TEXT);
    varptr += SIZE_TEXT;
    strncpy(problem_copyright,varptr,SIZE_COPYRIGHT);
}



/*===========================================================================*/
/* list all stored moves so far                                              */
/*===========================================================================*/
void HandleExport(void) {
    short i  = 0;
    short mv = move_count-moves_stored;
    FILE* fp;

    if (isArchived($(ticexpt))) return;   // <-- TIGCC 0.7 won't guess LOCK mode !!


    // check if there are any moves to export -> otherwise return immediately
    if (moves_stored == 0) return;


    // try to open export file for writing
    if (!(fp = fopen("ticexpt","w"))) return;

    fputs("[TI-Chess "TIC_VERSION_MAIN" Game]\n",fp);

    if (mv % 2 == 1) {
        fprintf(fp,"[%03d] ...... %s\n",mv+1,Move2Str((move_t*)&move_store[0],FALSE));
        i = 1;
    }

    for (;i<moves_stored;i+=2) {
        fprintf(fp,"[%03d] %s  ",mv+i+1,Move2Str((move_t*)&move_store[i],FALSE));
        if (i+1 < moves_stored) {
            fputs(Move2Str((move_t*)&move_store[i+1],FALSE),fp);
        }
        fputc('\n',fp);
    }

    fclose(fp);
}



//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: loadsave.c,v $
// Revision 1.5  2001/02/17 15:00:11  Thomas Nussbaumer
// changes due to new TIGCC version
//
// Revision 1.4  2000/12/19 13:55:29  Thomas Nussbaumer
// warnings stated by compiling with option -Wall fixed
//
// Revision 1.3  2000/12/09 15:37:23  Thomas Nussbaumer
// defaults loading and saving completely rewritten
//
// Revision 1.2  2000/08/12 15:31:13  Thomas Nussbaumer
// substitution keywords added
//
//
